#ifndef _FAT16_H_
#define _FAT16_H_

#include "type.h"


#define 	FAT_SIZE_BYTES		122880
#define 	FAT_SIZE_SECTORS 	240


#define 	BYTE_PER_SECTOR		(0x0B)
#define 	SECTOR_PER_CLUSTER	(0x0D)
#define 	RESERVE_SECTOR		(0x0E)
#define 	FAT_NUM				(0x10)
#define		ROOT_ENTRICE		(0x11)
#define 	SMALL_SECTOR		(0x13)
#define 	MEDIA_DESCRIPTOR	(0x15)
#define 	SECTOR_PER_FAT		(0x16)
#define 	SECTOR_PER_TRACK	(0x18)
#define 	HEAD_NUM			(0x1A)
#define 	HIDE_SECTOR			(0x1C)
#define 	LARGE_SECTOR		(0x20)
#define 	EXT_BOOT_SIGN		(0x26)
#define		VOLUME_SERIAL_NUM	(0x27)
#define 	VOLUME_LABLE		(0x2b)
#define 	FAT_TYPE			(0x36)


#define 	FILE_ATTR_READ_ONLY   	0x01
#define 	FILE_ATTR_HIDDEN 		0x02
#define 	FILE_ATTR_SYSTEM 		0x04
#define 	FILE_ATTR_SYSHID		0x06
#define 	FILE_ATTR_VOLUME_ID 	0x08
#define 	FILE_ATTR_DIRECTORY		0x10
#define 	FILE_ATTR_ARCHIVE  		0x20
#define 	FILE_ATTR_LFN_TEXT		0x0F
#define 	FILE_HEADER_BLANK		0x00
#define 	FILE_HEADER_DELETED		0xE5

#define 	EOS_POS				0x1FE
#define 	EOS_VALUE			0xAA55

#define 	MODE_BUF			0x03
#define 	MODE_FREEBUF		0x04
#define 	MODE_FREEFILE		0x08
#define 	MODE_READ			0x10
#define 	MODE_WRITE			0x20
#define 	MODE_RDWR			0x30
#define 	MODE_READING		0x40
#define 	MODE_WRITING		0x80
#define 	MODE_EOF			0x100
#define		MODE_ERR			0x200
#define 	MODE_UNGOT			0x400
#define		MODE_OPEN			0x800


#define 	FIND_FILE_MODE	(0)
#define		FIND_DIR_MODE   (1)
#define 	FIND_ALL_MODE   (2)


#define		O_APPEND	(0x01)
#define 	O_CREAT		(0x02)
#define		O_EXCL		(0x04)
#define 	O_RDONLY	(0x08)
#define 	O_RDWR		(0x10)
#define		O_TRUNC		(0x20)
#define		O_WRONLY	(0x40)


typedef struct
{
	BYTE short_name[11];
	BYTE type;
	BYTE reserve1;
	BYTE create_time_10ms;
	BYTE create_time[2];
	BYTE create_date[2];
	BYTE last_visit_date[2];
	BYTE res[2];
	BYTE last_edit_time[2];
	BYTE last_edit_date[2];
	BYTE first_cluster[2];
	BYTE filesize[4];
} file_label_t;

typedef struct
{
	BYTE num;
	BYTE name1[10];
	BYTE flag;
	BYTE reserve;
	BYTE chksum;
	BYTE name2[12];
	BYTE startcluster[2];
	BYTE name3[4];
} lfn_label_t;

typedef struct
{
	WORD	byte_per_sector;
	BYTE	sector_per_cluster;
	WORD	reserve_sector;
	BYTE	FATnum;
	WORD	root_entrice;
	DWORD	sector_num;
	DWORD	sector_per_FAT;
	char	label[11];
	char	file_system[8];
} BPB_t;


typedef struct
{
 	BPB_t 		bpb;
	DWORD		byte_per_cluster;
	DWORD 		fat_begin_lba;
	DWORD 		cluster_begin_lba;
	DWORD		total_cluster;
	DWORD		root_begin_lba;
} FS_t;

typedef struct
{
	WORD value;
	WORD offset;
} currentcluster_t;


 
typedef struct
{
	WORD	flag;
	WORD	startcluster;
	currentcluster_t currentcluster;
	WORD	curp;
	BYTE	filename[257];
	BYTE	type;
	DWORD	filesize;
	DWORD listp;
	WORD up_folder;
} FILE_t;

typedef struct
{
	FILE_t** list;
	DWORD    num;
} FILELIST_t;


typedef struct
{
    char path[15][257];
    int level;
} PATH_t;

BYTE* dummy;

// Variable declarations
FS_t *current_fs;									//Current file system structure

// Function Declarations

DWORD	fnFAT16_LBAofCluster(WORD Cluster_Number);
void 	fnFATChainInit();
void 	fnFAT16_init(FS_t *fs, DWORD offset_sector);
WORD 	return_WORD(BYTE* buff, DWORD offset);
WORD 	fnFAT16_FindNextCluster(WORD current_cluster);
currentcluster_t fnFAT16_SectorReader(currentcluster_t cluster, DWORD offset);
currentcluster_t fnFAT16_WriteSector(currentcluster_t cluster, DWORD offset);
void 	fnFAT16_UpdateFAT();
WORD 	fnFAT16_NextEmptyCluster();
void 	fnFAT16_EditFat(WORD Current_Cluster,WORD value);
void 	fnFAT16_EditFatChain(WORD Current_Cluster,WORD value);

int fnDetect_Disc(void);
int check_LFN(WORD recordoffset);
int check_SFEforLFN(WORD recordoffset,BYTE stringnum);
int check_INVALID(WORD recordoffset);
int check_SFN(WORD recordoffset);
BYTE fetch_LFN(WORD recordoffset,BYTE LFNIndex,BYTE *String);
BYTE fnGetChksum(BYTE *shortname);
void fnFindFile_LongName(DWORD sector_num,WORD item,DWORD up_folder,BYTE *String_buffer,WORD name_size,FILE_t *file);	
void fnFindFile_ShortName(DWORD sector_num,BYTE item,DWORD up_folder,FILE_t *file);
DWORD fnDetectDirectory(WORD startcluster);
FILELIST_t* fnListDirectory(WORD startcluster);
FILE_t *fnFindFile(WORD startcluster,BYTE *filename,BYTE mode);
DWORD fnReadFile(BYTE *buffer, BYTE size, DWORD count, FILE_t *fp);
void fnFreeFile(WORD startcluster);
FILE_t * fnOpenFile(FILE_t* fp,BYTE mode);
DWORD fnWriteFile(BYTE *buffer, BYTE size, DWORD count, FILE_t *fp);
BYTE fnDeleteFileLabel(FILE_t *fp);
WORD fnFindEmptyLabel(WORD startcluster,BYTE n);
BYTE fnGetShortNum(FILE_t *fp,BYTE *short_name);
BYTE fnUpdateFILE(FILE_t *fp);
FILE_t *fnCreateFile(WORD startcluster,BYTE *name);
void fnSearchFile(WORD startcluster,BYTE * name,BYTE level,BYTE *num,PATH_t* p);
currentcluster_t fnGetCurrentCluster(DWORD curp,WORD startcluster);

void lsDIR();
void printDIRlist(FILELIST_t *filelist1);
void printpath();
void init();
void GetFATInfo(void);
void changeDIR(char* dir_name);
void traverseupDIR();
void gotoROOTDIR();
void searchFILE(char* string);
//void eraseFILE(BYTE* string);
//void erase(BYTE* string,WORD startcluster);
void eraseDIR_recursive(WORD* stcluster);
//void eraseDIR(BYTE* string);
int readFILE(BYTE* source, BYTE* destination);
int createDIR(BYTE* string);
int createFILE(BYTE* string);
void clearcluster(WORD cluster_number);
int copyFILE(BYTE* source, BYTE* destination);
void erase_FILE_DIR(BYTE* string);
void erase(FILE_t* fp);
void init_SRU();



#endif