// Include files
#include "string.h"
#include "mmc_access.h"
#include "FAT16.h"
int ext_access=0;

section("seg_sdram")	BYTE fat_chain[FAT_SIZE_WORDS]; //array to store the FAT copied from external memory
WORD current_DIR=0;


	    
// Function to get the offset sector number of a cluster 
DWORD fnFAT16_LBAofCluster(WORD Cluster_Number)
{
    return ((current_fs->cluster_begin_lba + ((Cluster_Number-2)*current_fs->bpb.sector_per_cluster)));
}

// Function to initialize FAT chain from external memory to internal memory
void fnFATChainInit(FS_t *fs)
{
	ext_access=1;
    read_BufferSector(fat_chain,fs->fat_begin_lba,FAT_SIZE_SECTORS);//reads the whole FAT table from the start of the FAT sector to the FAT chain of internal memory
}

// Function to initialize the current file structure in the internal memory that 
// holds the important information about the File System
void fnFAT16_init(FS_t *fs, DWORD offset_sector)
{
	read_BufferSector(buffers.currentsector.u8data,offset_sector,1); //read the boot sector into the temporary buffer		

	//Initialize the current file system with the boot sector parameters
	WORD temp=read_WORD(BYTE_PER_SECTOR);
	//fs->bpb.byte_per_sector =	read_WORD(BYTE_PER_SECTOR);    
	fs->bpb.byte_per_sector =	temp;
	//read_WORD(BYTE_PER_SECTOR);    
	fs->bpb.sector_per_cluster = 	read_BYTE(SECTOR_PER_CLUSTER); 
	fs->bpb.reserve_sector =	read_WORD(RESERVE_SECTOR);     
	fs->bpb.FATnum =		read_BYTE(FAT_NUM);             
	fs->bpb.root_entrice = 		read_WORD(ROOT_ENTRICE);                
    fs->bpb.sector_num=             read_DWORD(LARGE_SECTOR);
	fs->bpb.sector_per_FAT=     	read_WORD(SECTOR_PER_FAT);
 
	DWORD num_rootsector=(fs->bpb.root_entrice*32+fs->bpb.byte_per_sector-1)/fs->bpb.byte_per_sector;
	int i=0; 
	for(i=0;i<11;i++)
	{
	     	 fs->bpb.label[i] =		read_BYTE(VOLUME_LABLE+i);
	}
	 
	for(i=0;i<8;i++)
	{
	     	 fs->bpb.file_system[i] =		read_BYTE(FAT_TYPE+i);
	}

	//Calculate the important offsets from the information read above
		 
	fs->byte_per_cluster= fs->bpb.byte_per_sector*fs->bpb.sector_per_cluster;//calculates the number of bytes per cluster
	
	fs->fat_begin_lba = offset_sector + fs->bpb.reserve_sector;	// calculates the sector offset for the starting sector of the FAT
	
	fs->cluster_begin_lba = fs->fat_begin_lba + (fs->bpb.FATnum * fs->bpb.sector_per_FAT)+num_rootsector;//Calculates the sector offset for the first data cluster
    
	fs->total_cluster=fs->bpb.sector_num/fs->bpb.sector_per_cluster;//calculates the total number of clusters
    
	fs->root_begin_lba=fs->fat_begin_lba+fs->bpb.sector_per_FAT*fs->bpb.FATnum;
	
	fnFATChainInit(fs);//Now initializes the FAT_chain structure with the FAT available in the external memory

	 
}

//Function to return a WORD value from any byte array from a specific location.
/*****************************************************************************/
WORD return_WORD(BYTE* buff, DWORD offset)
{
    WORD data;
    data=buff[offset];		//get the lower byte
    data+=buff[offset+1]<<8;//get the higher byte
    return data;
}    
/*****************************************************************************/



//Function to find the next cluster from the FAT table for the cuurent cluster value.
/***********************************************************************************/
WORD fnFAT16_FindNextCluster(WORD current_cluster)
{
	WORD nextcluster;
	WORD *cluster_buffer;
	

	if(current_cluster>current_fs->total_cluster)	//check if the cluster value given is greater than the maximum clusters available
		return 0xFFFF;

	nextcluster = return_WORD(fat_chain,current_cluster*2);//get the FAT entry for the cluster offset given
	
	if((nextcluster&0x0FFFF)==0x0FFFF)//If the FAT entry denotes the end of file, return 0xFFFFFFFF
	{
			return 0xFFFF;
	}
	else
	{						
		return nextcluster;			//If the above conditions are not true, return the next cluster number
	}

}
/*********************************************************************************/


//Function to read an offset sector of a file given the current cluster offset for that file.
/*****************************************************************************/

currentcluster_t fnFAT16_SectorReader(currentcluster_t cluster, DWORD offset)
{
 	  WORD SectortoRead = 0;
	  WORD ClustertoRead = 0;
	  WORD ClusterChain = 0;
	  
	  WORD sector_per_cluster=current_fs->bpb.sector_per_cluster;//get the sectors per cluster value from the information available in the file system
	  	  
	  if(cluster.value==0xFFFF)	//Check if it is the end of the file	
	  	return cluster;
	  
	  ClusterChain = cluster.value; //Intialize a variable to keep track of the cluster chain

	  ClustertoRead = offset / sector_per_cluster;	 //Get the cluster offset which is to be read for a file
	  SectortoRead = offset - (ClustertoRead*sector_per_cluster);//Get the sector offset within that cluster
	  
	  int i;
	  for (i=cluster.offset; i<ClustertoRead; i++)	//Go on tracing the chain till it reaches the target cluster
	  {
	  	  ClusterChain = fnFAT16_FindNextCluster(ClusterChain);  
	  }

	  cluster.value=ClusterChain;		//Intialize the current cluster
	  
	  cluster.offset=ClustertoRead;	
	  
	  if (ClusterChain==0xFFFF)  // Check if the value read at last is the end of the file
	  	  	return cluster;

	  //read target sector
	  read_BufferSector(buffers.currentsector.u8data,fnFAT16_LBAofCluster(ClusterChain)+SectortoRead,1);

	  return cluster;
}
/******************************************************************************************************************/



//Function to write an offset sector of a file given the current cluster offset for that file.
/********************************************************************************************/

currentcluster_t fnFAT16_WriteSector(currentcluster_t cluster, DWORD offset)
{
 	  WORD SectortoWrite = 0;
	  WORD ClustertoWrite = 0;
	  WORD ClusterChain = 0;
	  
	  int i;
	  
	  if(cluster.value==0xFFFF)			//Check if it is the end of the file
	  	return cluster;

	  ClusterChain = cluster.value;			//Intialize a variable to keep track of the cluster chain

	
	  ClustertoWrite = offset / current_fs->bpb.sector_per_cluster;	  //Get the cluster offset which is to be written for a file
	  SectortoWrite = offset - (ClustertoWrite*current_fs->bpb.sector_per_cluster);//Get the sector offset within that cluster
	  
	 
	  for (i=cluster.offset; i<ClustertoWrite; i++)		//Go on tracing the chain till it reaches the target cluster
	  	  ClusterChain = fnFAT16_FindNextCluster(ClusterChain);

	  //register current cluster
	  
	  cluster.value=ClusterChain;
	  
	  cluster.offset=ClustertoWrite;
	  
	  if (ClusterChain==0xFFFF) 	//Return if reach the end of the file
	  { 
	  	  return cluster;
	  }
	  
	  //write target sector
	  write_BufferSector(buffers.currentsector.u8data,fnFAT16_LBAofCluster(ClusterChain)+SectortoWrite,1);
	  
	  return cluster;
}
/***************************************************************************************************************************************/


//Function to update the FAT of the external memory from internal memory 
/************************************************************************/

void fnFAT16_UpdateFAT()
{
				
		write_BufferSector(fat_chain,current_fs->fat_begin_lba,FAT_SIZE_SECTORS);
	
		write_BufferSector(fat_chain,current_fs->fat_begin_lba+current_fs->bpb.sector_per_FAT,FAT_SIZE_SECTORS);
	
}
/************************************************************************************************/


//Function to find the next empty cluster 
/*****************************************/

WORD fnFAT16_NextEmptyCluster()
{
    int i=0;
	for(i=1;i<(current_fs->total_cluster);i++)	//Go on tracing the FAT table till an empty cluster is found
	{
	    WORD temp=return_WORD(fat_chain,i*2);
	    
	    if(temp==0x0000)
	    	return i;
	}
	
	return 0xFFFF;
}
/************************************************/


//Function to edit FAT for a particular cluster entry
/******************************************************/

void fnFAT16_EditFat(WORD Current_Cluster,WORD value)
{
	
    DWORD temp=Current_Cluster*2;
    WORD sector_offset= (temp/current_fs->bpb.byte_per_sector);
    WORD u16position =temp-sector_offset*512;
    BYTE* buffer=(BYTE*)(fat_chain);
    
    buffer+=sector_offset*512;
    
    buffer[u16position]=value&0x000000FF;
    buffer[u16position+1]=(value&0x0000FF00)>>8;
    write_WORD(u16position,value);
    
    write_BufferSector(buffer,current_fs->fat_begin_lba+sector_offset,1);
    
}
/**************************************************************************/


//Function to edit FAT for a particular cluster entry only in internal memory
/***************************************************************************/
void fnFAT16_EditFatChain(WORD Current_Cluster,WORD value)
{
	
    DWORD temp=Current_Cluster*2;
    WORD sector_offset= (temp/current_fs->bpb.byte_per_sector);
    WORD u16position =temp-sector_offset*512;
    BYTE* buffer=(BYTE*)(fat_chain);
    
    buffer+=sector_offset*512;
        
    buffer[u16position]=value&0x000000FF;
    buffer[u16position+1]=(value&0x0000FF00)>>8;
            
}